/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "build.h"
#include "mapLlhd.h"
#include "mapPool.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define NUM_MARKERS_INIT 6
#define BUILD_WIDTH 3
#define BUILD_WIDTH_SEED 5

void buildMapPerfect(map_t *map, pedigree_t *ped, int numMarks,
                     marker_t **marks, int split);
void buildMapPoolPerfect(mapPool_t *, pedigree_t *ped, int numMarks,
                         marker_t **marks, int split);

int incrPermutation(int *perm, int n);

double computeInformness(const marker_t *m, int numInds);

/* This is used for passing the number of individuals to compareMarkPtr. */
static int numInds;

int compareMarkPtr(const void *p1, const void *p2)
{
    marker_t    *m1 = *(marker_t *const *)p1;
    marker_t    *m2 = *(marker_t *const *)p2;
    double      inf1 = computeInformness(m1, numInds);
    double      inf2 = computeInformness(m2, numInds);
    
    if (inf1 > inf2)
        return -1;
    else if (inf1 < inf2)
        return 1;
    else
        return 0;
}

/* Builds a reasonable map from the given set of markers. It initially tries
   all non-equivalent orders for a subset of the markers and then adds markers
   by trying all possible insertion points.
 */
void buildMap(map_t *map, pedigree_t *ped, int numMarks, marker_t **marks,
              int split)
{
    int         i, j, k;
    map_t       curMap;
    mapPool_t   lastPool, curPool;
    marker_t    **marksCopy;
    
    /* If numMarks < NUM_MARKERS_INIT, then the standard code path will crash,
       If numMarks is NUM_MARKERS_INIT, then the standard path will work, but
       the following is simpler and no less efficient.
     */
    if (numMarks <= NUM_MARKERS_INIT) {
        buildMapPerfect(map, ped, numMarks, marks, split);
        return;
    }
    
    /* order markers. */
    marksCopy = (marker_t**)malloc(sizeof(marker_t *) * numMarks);
    memcpy(marksCopy, marks, numMarks * sizeof(marker_t *));
    numInds = ped->numInds;
    qsort(marksCopy, numMarks, sizeof(marker_t *), compareMarkPtr);
    
        /* The + 1 is so that the shuffle step of inserting a new marker into
           curMap does not go beyond the end of the buffer. */
    allocMap(&curMap, ped, numMarks + 1, split);
    allocMapPool(&lastPool, BUILD_WIDTH);
    allocMapPool(&curPool, BUILD_WIDTH_SEED);
    
    buildMapPoolPerfect(&curPool, ped, NUM_MARKERS_INIT, marksCopy, split);
    
    for (i = NUM_MARKERS_INIT; i < numMarks; ++i) {
        marker_t    *newMarker = marksCopy[i];
        
        swapMapPools(&curPool, &lastPool);
        clearMapPool(&curPool);
        curPool.capacity = BUILD_WIDTH;
        
        for (k = 0; k < lastPool.size; ++k) {
            copyMap(&curMap, lastPool.maps + k);
            curMap.numMarks = i + 1;
            
            for (j = i; j >= 0; --j) {
                if (j == i - 1) {
                    curMap.order[j + 1] = curMap.order[j - 1];
                    curMap.order[j - 1] = curMap.order[j];
                } else if (j == 0) {
                    curMap.order[1] = curMap.order[2];
                    curMap.order[2] = curMap.order[0];
                } else
                    curMap.order[j + 1] = curMap.order[j];
                curMap.order[j] = newMarker;
                
                computeMapLlhd(&curMap);
                insertMapPool(&curPool, &curMap);
            }
        }
    }
    
    freeMap(&curMap);
    allocMap(map, ped, numMarks, split);
    copyMap(map, bestMapPool(&curPool));
    freeMapPool(&curPool);
    freeMapPool(&lastPool);
}

/* Constructs the map with the highest log likelihood by trying all n!/8
   possible orders.
 */
void buildMapPerfect(map_t *map, pedigree_t *ped, int numMarks,
                     marker_t **marks, int split)
{
    mapPool_t   pool;
    
    allocMapPool(&pool, 1);
    buildMapPoolPerfect(&pool, ped, numMarks, marks, split);
    *map = *pool.maps;
    pool.size = 0;
    freeMapPool(&pool);
}

/* This tries all possible orders, and keeps the best in the pool.
 */
void buildMapPoolPerfect(mapPool_t *pool, pedigree_t *ped, int numMarks,
                         marker_t **marks, int split)
{
    int         *perm;
    map_t       curMap;
    int         i;
    
    allocMap(&curMap, ped, numMarks, split);
    
    /* For numMarks < 4, there all possible orderings are equivalent, so the
       order the markers are entered is as good as any.  Furthermore, the loop
       below won't work for numMarks < 3, and for numMarks == 3, the following
       is equivalent, simpler, and more efficient.
     */
    if (numMarks < 4) {
        for (i = 0; i < numMarks; ++i)
            curMap.order[i] = marks[i];
        computeMapLlhd(&curMap);
        
        insertMapPool(pool, &curMap);
        freeMap(&curMap);
        return;
    }
    
    perm = (int *)malloc(numMarks * sizeof(int));
    for (i = 0; i < numMarks; ++i)
        perm[i] = i;
    
    do {
        /* This avoids repeating equivalent orders. */
        if (perm[0] > perm[numMarks - 1] || perm[0] > perm[1]
              || perm[numMarks - 1] > perm[numMarks - 2])
            continue;
        
        for (i = 0; i < numMarks; ++i)
            curMap.order[i] = marks[perm[i]];
        computeMapLlhd(&curMap);
        
        insertMapPool(pool, &curMap);
    } while(incrPermutation(perm, numMarks));
    
    free(perm);
    freeMap(&curMap);
}

/* perm is a vector of length n consisting of the numbers 0...n-1, each exactly
   once. Beginning with the trivial permutation, i.e. 0 through n-1 in that
   order, this function cycles through all n! permutations. The return value
   is FALSE when the end of the cycle is reached. n must be at least 2.
 */
int incrPermutation(int *perm, int n)
{
    int     i, j, k;
    int     tmp;
    
    for (i = n - 2; perm[i] > perm[i + 1]; --i) {
        if (i == 0)
            return FALSE;
    }
    
    for (j = i + 1; j < n && perm[j] > perm[i]; ++j)
        ;
    
    tmp = perm[i];
    perm[i] = perm[j - 1];
    perm[j - 1] = tmp;
    
    for (j = i + 1, k = n - 1; j < k; ++j, --k) {
        tmp = perm[j];
        perm[j] = perm[k];
        perm[k] = tmp;
    }
    
    return TRUE;
}

/* computes the information content of a given marker. */
double computeInformness(const marker_t *m, int numInds)
{
    double  result = 0.0;
    double  inform[16];
    int     i;
    
    for (i = 0; i < 16; ++i) {
        int     j, numBits = 0;
        
        for (j = 1; j < 16; j *= 2) {
            if (i & j)
                ++numBits;
        }
        
        inform[i] = -log(numBits / 4.0);
    }
    
    for (i = 0; i < numInds; ++i)
        result += inform[m->obs[i]];
    
    return result;
}
