/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 * Copyright 2010 Dustin Cartwright
 *
 * All rights reserved.
 */

package com.myriad.tmap.compute;

import java.io.File;
import java.util.regex.*;
import javax.swing.JOptionPane;

/** This class provides a number of computational methods on a pedigree. Most of
    the work is done by native methods written in C. The Pedigree class provides
    wrappers for these functions.
 */
public class Pedigree {
    /** Creates a new class from the pedigree data found in the specified file.
     */
    public Pedigree(String aPath) throws Exception
    {
        ped = readPedigreeFile(aPath);
        if (ped == 0)
            throw new Exception("Couldn't read pedigree file.");
        filename = (new File(aPath)).getName();
    }
    
    /** Frees the memory held by the native methods. */
    protected void finalize()
    {
        if (ped != 0)
            freePed(ped);
    }

    public String getFilename()
    {
        return filename;
    }

    public Swaps phasing()
    {
        long    swapsPtr =  phasingNative(ped);
        if (swapsPtr != 0) {
            return new Swaps(swapsPtr);
        } else {
            return null;
        }
    }
    
    /** Computes all the two point LOD scores and distances. */
    public TwoPointData compute2Pts()
    {
        int         numMarks = getNumMarkers(ped);
        String[]    names = new String[numMarks];
        double[][]  lods = new double[numMarks][];
        double[][]  dists = new double[numMarks][];
        
        compute2PtsNative(ped, names, lods, dists);
        return new TwoPointData(names, lods, dists);
    }
    
    /** Computes the map (i.e. distances and errors) for the given marker
        order.
     */
    public Map computeMap(String[] markerNames, boolean split)
    {
        return makeMap(markerNames, split, MAKE_MAP_COMPUTE_DISTS);
    }
    
    /** Returns a set of markers from the given file. */
    public String[] getOrder(String orderPath)
    {
        String[]    markers = getOrderNative(ped, orderPath);

        if (markers == null) {
            JOptionPane.showMessageDialog(null,
                                          "Couldn't read linkage group file",
                                          "Linkage group file",
                                          JOptionPane.ERROR_MESSAGE);
        }
        return markers;
    }
    
    /** Builds an initial map order based on the markers in the markerNames
        array. On return, markerNames has been altered to reflect the new
        order of the markers.
     */
    public Map build(String[] markerNames, boolean split)
    {
        return makeMap(markerNames, split, MAKE_MAP_BUILD);
    }

    /** Improves the map order indicated by markerNames. On return, the
        markerNames array has been altered to reflect the new order.
     */
    public Map improve(String[] markerNames, boolean split)
    {
        return makeMap(markerNames, split, MAKE_MAP_IMPROVE);
    }

    public void writePed(File file)
    {
        writePedNative(ped, file.getAbsolutePath());
    }
    
    /** Outputs the map for a given marker order to File.
        @param markerNames an array containg the markers which make up the map
        @param split whether the maternal and paternal distances should be
        computed independently
        @param file the file which will contain the output
     */
    public void writeMap(String[] markerNames, boolean split, File file)
    {
        writeMapNative(ped, markerNames, split, file.getAbsolutePath());
    }
    
    /** This is a wrapper function around makeMapNative. It's main purpose is to
        give arrays to the native function and then to package these into the
        Map object.
     */
    private Map makeMap(String[] markerNames, boolean split, int type)
    {
        if (markerNames.length > 0) {
            double[]    distsMoth = new double[markerNames.length - 1];
            double[]    distsFath = new double[markerNames.length - 1];
            double[]    errorsMoth = new double[markerNames.length];
            double[]    errorsFath = new double[markerNames.length];
            Map         map;
            Thread      curThr = Thread.currentThread();
            int         pri = curThr.getPriority();
            
            /* Since this may take a while, we decrease our priority so that
               other work may continue. Compute dists is not affected, since it
               is usually fast and should be interactive.
             */
            if (type != MAKE_MAP_COMPUTE_DISTS)
                curThr.setPriority(Thread.MIN_PRIORITY);
            
            makeMapNative(ped, markerNames, split, distsMoth, distsFath,
                          errorsMoth, errorsFath, type);
            map = buildMap(markerNames, distsMoth, distsFath, errorsMoth,
                           errorsFath);
            
            if (type != MAKE_MAP_COMPUTE_DISTS)
                curThr.setPriority(pri);
            return map;
        } else
            return new Map(new MarkerPosition[0]);
    }
    
    /** Builds a map class from the given data. The native classes return maps
        in the form of separate double arrays for the distances and the errors.
        This translates this data to a more Java-like structure.
     */
    private Map buildMap(String[] markerNames, double[] distsMoth,
                         double[] distsFath, double[] errorsMoth,
                         double[] errorsFath)
    {
        MarkerPosition[]    markers = new MarkerPosition[markerNames.length];
        double              totalDistMoth = 0.0;
        double              totalDistFath = 0.0;
        
        for (int i = 0; i < markerNames.length; ++i) {
            double  distMoth, distFath;
            
            if (i < markerNames.length - 1) {
                distMoth = distsMoth[i];
                distFath = distsFath[i];
            } else
                distMoth = distFath = 0.0;
            
            markers[i] = new MarkerPosition(markerNames[i],
                                            totalDistMoth, totalDistFath,
                                            errorsMoth[i], errorsFath[i]);
            totalDistMoth += distMoth;
            totalDistFath += distFath;
        }
        
        return new Map(markers);
    }
    
    private native long     readPedigreeFile(String path);
    private native void     freePed(long ped);
    private native int      getNumMarkers(long ped);
    private native long     phasingNative(long ped);
    private native void     compute2PtsNative(long ped, String[] names,
                                              double[][] lods,
                                              double [][] dists);
    private native void     writePedNative(long ped, String path);
    private native String[] getOrderNative(long ped, String orderPath);
    private native void     writeMapNative(long ped, String[] markerNames,
                                           boolean split, String path);
    private native void     makeMapNative(long ped, String[] markerNames,
                                          boolean split,
                                          double[] distsMoth,
                                          double[] distsFath,
                                          double[] errorMoth,
                                          double[] errorFath, int type);
     
    /* Functions for the Swaps class. These are here for two reasons:
         - Only one automatically generated header file for the C code
         - The Swaps class doesn't seem to see the JNI symbols.
     */
    private native void     writeJoinMapFileNative(long ped, long swapsPtr,
                                               String path);
    private native void     freeSwaps(long swapsPtr);
    
    public class Swaps {
        private Swaps(long aSwapsPtr)
        {
            swapsPtr = aSwapsPtr;
        }
        
        protected void finalize()
        {
            freeSwaps(swapsPtr);
        }
        
        public void writeJoinMapFile(File file)
        {
            writeJoinMapFileNative(ped, swapsPtr, file.getAbsolutePath());
        }
        
        private long swapsPtr;
    }
            
    
    private long    ped;
    private String  filename;
    
    private final static int    MAKE_MAP_COMPUTE_DISTS = 0;
    private final static int    MAKE_MAP_BUILD = 1;
    private final static int    MAKE_MAP_IMPROVE = 2;
    
    /** This static code section tries to load the native code parts of the
        class. First, it tries using loadLibrary and the default library path.
        Then, it tries to find the path of the jar file which was used to launch
        the, and looks in that directory.
     */
    static {
        boolean succeeded = false;
        String  err_msg = "";
        
        try {
            System.loadLibrary("TmapJni");
            succeeded = true;
        } catch (UnsatisfiedLinkError e) {
            err_msg = e.getMessage();
        }
        
        if (!succeeded) {
            String      cp = System.getProperty("java.class.path");
            String[]    libNames = {"TmapJni.dll", "libTmapJni.so"};
                /* Separates the path component from the jar file name. Assumes:
                     1. The directory separater is not in \w
                     2. The JAR file name is in \w
                     3. There is only one path in the class path. */
            Matcher     matcher = Pattern.compile("(.*?)\\w*.jar").matcher(cp);
            
            if (matcher.matches()) {
                String      path = matcher.group(1);
                
                for (int i = 0; i < libNames.length; ++i) {
                    try {
                        System.load(path + libNames[i]);
                        succeeded = true;
                        break;
                    } catch(UnsatisfiedLinkError e) {}
                }
            }
        }
        
        if (!succeeded) {
            JOptionPane.showMessageDialog(null,
                                          "Couldn't find TmapJni library: "
                                            + err_msg,
                                          "No TmapJni library",
                                          JOptionPane.ERROR_MESSAGE);
            System.exit(0);
        }
    }
}
