/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

package com.myriad.tmap.grouping;

import java.awt.*;
import java.util.ArrayList;

/** Represents a linkage group which exists and a lod score intermediate
    between the minimum and the maximum. It consists of two linkage groups and
    the lod score at which they are merged.
 */
class LinkGroupMerge extends LinkGroupNode {
    public LinkGroupMerge(LinkGroupNode n1, LinkGroupNode n2, double aLodScore)
    {
        child1 = n1;
        child2 = n2;
        n1.parent = this;
        n2.parent = this;
        lodScore = aLodScore;
        divider = 0;
    }
    
    public int position(int upperBound, double minLod)
    {
        divider = child1.position(upperBound, minLod);
        return child2.position(divider + (minLod < lodScore ? 0 : clumpGap),
                               minLod);
    }
    
    public Point paint(Graphics2D g, DisplayParameters params, int top,
                       int bottom)
    {
        Point   p1, p2;
        Point   p = new Point();
        
        if (divider >= top)
            p1 = child1.paint(g, params, top, bottom);
        else
            /* The x-coordinate is not close to correct here, but it doesn't
               matter because it won't be visible anyways. */
            p1 = new Point(0, divider);
        if (divider <= bottom)
            p2 = child2.paint(g, params, top, bottom);
        else
            p2 = new Point(0, divider);
        
        if (lodScore > params.maxLod)
            p.x = params.markerX;
        else
            p.x = (int) (params.scale * (lodScore - params.minLod));
        p.y = (p1.y + p2.y) / 2;
        
        if (p.x > 0) {
            setPaintColor(g);
            g.drawLine(p1.x, p1.y, p.x, p1.y);
            g.drawLine(p2.x, p2.y, p.x, p2.y);
            g.drawLine(p.x, p1.y, p.x, p2.y);
        }
        
        return p;
    }
    
    public int getPreferredHeight(double minLod)
    {
        return child1.getPreferredHeight(minLod)
                + child2.getPreferredHeight(minLod)
                + (minLod < lodScore ? 0 : clumpGap);
    }
    
    public int getMaxMarkerWidth(Graphics2D g)
    {
        int     width1 = child1.getMaxMarkerWidth(g);
        int     width2 = child2.getMaxMarkerWidth(g);
        
        return (width1 > width2) ? width1 : width2;
    }
    
    public LinkGroupNode findNodeForPoint(Point p, DisplayParameters params)
    {
        if (p.x <= (int) (params.scale * (lodScore - params.minLod)))
            return this;
        else {
            if (p.y < divider)
                return child1.findNodeForPoint(p, params);
            else if (p.y > divider)
                return child2.findNodeForPoint(p, params);
            else
                return null;
        }
    }
    
    public Marker findMarkerByName(String name)
    {
        Marker  mark = child1.findMarkerByName(name);
        
        if (mark != null)
            return mark;
        else
            return child2.findMarkerByName(name);
    }
    
    void setSelected(boolean sel)
    {
        child1.setSelected(sel);
        child2.setSelected(sel);
    }
    
    int getStatus()
    {
        return child1.getStatus() & child2.getStatus();
    }
    
    public LinkGroupNode cluster(double maxLod)
    {
        if (lodScore < maxLod)
            return new LinkGroupMerge(child1.cluster(maxLod),
                                      child2.cluster(maxLod), lodScore);
        else
            return new LinkGroupCluster(getMarkerArray(false));
    }
    
    protected void getMarkerList(ArrayList acc, boolean sel)
    {
        child1.getMarkerList(acc, sel);
        child2.getMarkerList(acc, sel);
    }
    
    private LinkGroupNode   child1;
    private LinkGroupNode   child2;
    private double          lodScore;
    private int             divider;
        /** This is the extra space added between groups which are separated
            even at the minimum lod score. */
    private static final int    clumpGap = 7;
}
