/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 * Copyright 2010 Dustin Cartwright
 *
 * All rights reserved.
 */

package com.myriad.tmap.improve;

import com.myriad.tmap.compute.*;
import java.awt.EventQueue;

/** This is used to create a worker thread which computes maps from marker
    orders. The marker orders can be submitted and the results will be returned
    asynchronously to the designated receiver. The returned results are always
    returned in the same order as they are given, but some may be skipped. The
    calls to mapResult will always be in the event-dispatching thread or from
    the thread which calls computeMap (probably also the event-dispatching
    thread).

    Note that it may call mapResult while the PedigreeThread object is locked.
    Therefore, any lock which is held both before calling computeMap and in
    mapResult may cause a deadlock.
 */
class PedigreeThread implements Runnable {
    PedigreeThread(MapReceivable aRecv, Pedigree aPed)
    {
        recv = aRecv;
        ped = aPed;
        order = null;
        running = false;
        relevant = true;
        cache = new MapCache();
    }
    
    /** This puts a given map on the queue to be computed. If there is no other
        computation in process, the worker thread will begin computing it
        immediately. Otherwise, it will do so after it has finished with the
        current computation.
     */
    public synchronized void computeMap(String [] aOrder, boolean aSplit)
    {
        Map     cachedMap = cache.getMap(aOrder, aSplit);

        if (cachedMap != null) {
            order = null;
            relevant = false;

            recv.mapResult(cachedMap);
        } else {
            order = aOrder;
            split = aSplit;

            if (!running) {
                (new Thread(this)).start();
                running = true;
            }
        }
    }

    /** The working thread loop. It computes maps so long as there is a new map
     * in the buffer.
     */
    public void run()
    {
        for(;;) {
            String[]    computeOrder;
            boolean     computeSplit;
            
            synchronized (this) {
                computeOrder = order;
                if (computeOrder == null) {
                    running = false;
                    return;
                }
                computeSplit = split;
                order = null;
                relevant = true;
            }
            
            Map map = ped.computeMap(computeOrder, computeSplit);
            cache.cacheMap(computeOrder, computeSplit, map);
            synchronized (this) {
                if (relevant)
                    EventQueue.invokeLater(new ResultTask(map));
            }
        }
    }

    private class ResultTask implements Runnable {
        ResultTask (Map aMap) {
            map = aMap;
        }

        public void run() {
            recv.mapResult(map);
        }

        Map     map;
    }

    private MapReceivable   recv;
    private Pedigree        ped;
    private String[]        order;
    private boolean         split;
    private boolean         running;
    private boolean         relevant; /* Is the current Pedigree.computeMap
                                       * computation more recent than any
                                       * call to mapResult? */
    private MapCache        cache;
};
