/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "genmap.h"
#include "input.h"
#include "inputOrder.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int compareGroups(marker_t **markers1, marker_t **markers2, int numMarkers,
                  int numInd, int *mask);
int compareGroupsFixedPhases(marker_t **markers1, marker_t **markers2,
                             int numMarkers, int numInd, int *mask);
void swapMaternalPhases(marker_t **markers, int numMarkers, int numInd);
void swapPaternalPhases(marker_t **markers, int numMarkers, int numInd);

/* This program compares a single linkage group in two different pedigrees to
   see if the markers have identical genotypes.
 */
int main(int argc, char **argv)
{
    pedigree_t  ped1, ped2;
    marker_t    **markers1, **markers2;
    int         numMarkers1, numMarkers2;
    
    if (argc != 4) {
        fprintf(stderr, "usage: %s <ped> <ped> <linkage group file>\n", *argv);
        return EXIT_FAILURE;
    }
    
    if (!readFile(&ped1, argv[1]))
        return EXIT_FAILURE;
    if (!readFile(&ped2, argv[2]))
        return EXIT_FAILURE;
    
    markers1 = readOrderFile(&ped1, &numMarkers1, argv[3]);
    markers2 = readOrderFile(&ped2, &numMarkers2, argv[3]);
    if (markers1 == NULL || markers2 == NULL)
        return EXIT_FAILURE;
    
    if (ped1.numInds != ped2.numInds)
        printf("different numbers of individuals\n");
    else if (numMarkers1 != numMarkers2)
        printf("different number of markers\n");
    else {
        int numDiff;
        int *mask = (int *)malloc(numMarkers1 * sizeof(int));
        
        numDiff = compareGroups(markers1, markers2, numMarkers1, ped1.numInds,
                                mask);
        if (numDiff == 0)
            printf("equivalent\n");
        else {
            int     i;
            
            printf("not equivalent in %d markers:\n", numDiff);
            for (i = 0; i < numMarkers1; ++i) {
                if (mask[i])
                    printf("%s\n", markers1[i]->name);
            }
        }
        
        free(mask);
    }
    
    free(markers1);
    free(markers2);
    freePed(&ped1);
    freePed(&ped2);
    return EXIT_SUCCESS;
}

/* Compares the markers from two pedigrees to each other. The phases are not
   assumed to be consistent between the two groups. In other words, it tries
   all possible phases for one group of markers (keeping the relative phases
   constant) and chooses the best. It returns the number of different markers
   and sets mask[i] to be true if the ith marker is different.
 */
int compareGroups(marker_t **markers1, marker_t **markers2, int numMarkers,
                  int numInd, int *mask)
{
    int     bestNum = numMarkers + 1;
    int     *tmpMask = (int *)malloc(numMarkers * sizeof(int));
    int     i;
    
    for (i = 0; i < 4; ++i) {
        int     num;
        
        num = compareGroupsFixedPhases(markers1, markers2, numMarkers, numInd,
                                       tmpMask);
        if (num < bestNum) {
            bestNum = num;
            if (mask)
                memcpy(mask, tmpMask, numMarkers * sizeof(int));
        }
        if (i % 2)
            swapPaternalPhases(markers1, numMarkers, numInd);
        else
            swapMaternalPhases(markers1, numMarkers, numInd);
    }
    
    free(tmpMask);
    return bestNum;
}

/* Compares the markers in two different pedigrees directly. Simply compares
   all the genotypes to see if they're equal. Returns the number of markers
   which differ and sets mask[i] if the ith markers differ.
 */
int compareGroupsFixedPhases(marker_t **markers1, marker_t **markers2,
                             int numMarkers, int numInd, int *mask)
{
    int     i, j;
    int     numDiffs = 0;
    
    for (i = 0; i < numMarkers; ++i) {
        mask[i] = FALSE;
        for (j = 0; j < numInd; ++j) {
            if (markers1[i]->obs[j] != markers2[i]->obs[j]) {
                ++numDiffs;
                if (mask)
                    mask[i] = TRUE;
                break;
            }
        }
    }
    return numDiffs;
}

/* Swaps the maternal phases of all markers. */
void swapMaternalPhases(marker_t **markers, int numMarkers, int numInd)
{
    int     i, j;
    
    for (i = 0; i < numMarkers; ++i) {
        for (j = 0; j < numInd; ++j) {
            obs_t   obs = markers[i]->obs[j];
            markers[i]->obs[j] = ((obs & 0xa) >> 1) | ((obs & 0x5) << 1);
        }
    }
}

/* Swaps the paternal phases of all markers. */
void swapPaternalPhases(marker_t **markers, int numMarkers, int numInd)
{
    int     i, j;
    
    for (i = 0; i < numMarkers; ++i) {
        for (j = 0; j < numInd; ++j) {
            obs_t   obs = markers[i]->obs[j];
            markers[i]->obs[j] = (obs >> 2) | ((obs & 0x3) << 2);
        }
    }
}


