/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "genmap.h"
#include "input.h"
#include "inputOrder.h"
#include "map.h"
#include "mapLlhd.h"
#include "ped.h"
#include "improve.h"
#include "build.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

void print_usage(const char *progname);

/* Takes a list of markes and generates a genetic map from them. The flags can
   be used to specify the combination of steps used to generate the map.
 */
int main(int argc, char **argv)
{
    int         c;
    int         buildFlag = FALSE;
    int         improveFlag = FALSE;
    int         normalizeFlag = TRUE;
    int         logLlhdFlag = FALSE;
    int         splitFlag = FALSE;
    char        *pedFile, *orderFile;
    pedigree_t  ped;
    marker_t    **markers;
    int         numMarks;
    map_t       map;
    
    while ((c = getopt(argc, argv, "bhinls")) != EOF) {
        switch (c) {
            case 'b':
                buildFlag = TRUE;
                break;
            case 'h':
                print_usage(*argv);
                return EXIT_SUCCESS;
            case 'i':
                improveFlag = TRUE;
                break;
            case 'n':
                normalizeFlag = FALSE;
                break;
            case 'l':
                logLlhdFlag = TRUE;
                break;
            case 's':
                splitFlag = TRUE;
                break;
        }
    }
    
    if (argc - optind < 1) {
        fprintf(stderr, "%s: Too few arguments\n", *argv);
        print_usage(*argv);
        return EXIT_FAILURE;
    } else {
        pedFile = argv[optind];
        
        if (argc - optind == 1)
            orderFile = NULL;
        else if (argc - optind == 2)
            orderFile = argv[optind + 1];
        else {
            fprintf(stderr, "%s: Too many arguments\n", *argv);
            print_usage(*argv);
            return EXIT_FAILURE;
        }
    }
    
    if (!readFile(&ped, pedFile)) {
        printf("failed\n");
        return EXIT_FAILURE;
    }
    
    if (orderFile)
        markers = readOrderFile(&ped, &numMarks, orderFile);
    else
        markers = readOrder(&ped, &numMarks, stdin);
    if (markers == NULL) {
        freePed(&ped);
        return EXIT_FAILURE;
    }
    
    if (buildFlag) {
        buildMap(&map, &ped, numMarks, markers, splitFlag);
    } else {
        allocMap(&map, &ped, numMarks, splitFlag);
        memcpy(map.order, markers, numMarks * sizeof(marker_t *));
        computeMapLlhd(&map);
    }
    
    free(markers);
    
    if (improveFlag)
        improveMap(&map);
    
    if (normalizeFlag)
        normalizeMap(&map);
    
    if (logLlhdFlag)
        printf("log likelihood = %.2f\n", map.logLlhd);
    else
        printMap(stdout, &map);
    
    freeMap(&map);
    freePed(&ped);
    return EXIT_SUCCESS;
}

void print_usage(const char *progname)
{
    fprintf(stderr, "usage: %s [-bhinls] <pedigree> [<order file>]\n", progname);
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "    -b Build map using markers in <order file>, but ignoring their order\n");
    fprintf(stderr, "    -h Print this help message\n");
    fprintf(stderr, "    -i Improve map order, starting from given order\n");
    fprintf(stderr, "    -n DON'T normalize map by changing the order of terminal markers\n");
    fprintf(stderr, "    -l Print log likelihood instead of full map\n");
    fprintf(stderr, "    -s Split: allow different recombinations rates for mother and father\n");
    fprintf(stderr, "By default, tmap just computes the distances for the given order. Use the -b\n");
    fprintf(stderr, "and/or -i options to try to find the correct order. If no order file is given,\n");
    fprintf(stderr, "tmap will read the marker list from standard input\n");
}
